package acqConstraints;

import java.util.ArrayList;
import java.util.List;

import org.chocosolver.solver.Model;
import org.chocosolver.solver.constraints.Constraint;
import org.chocosolver.solver.variables.IntVar;

import vocabulary.AcqVariable;

/**
 * An acquisition constraint dedicated to arithmetic operations.
 * <br/>
 * There are three available definitions:
 * <li>
 * <ul>VAR op1 CSTE,</ul>
 * <ul>VAR op1 VAR,</ul>
 * <ul>VAR op1 VAR op2 CSTE</ul>
 * </li>
 * where VAR is a variable, CSTE a constant and op is an operator among {"=", "!=","<", ">", "<=, ">="} or{"+", "-"}.
 *
 * @author LAZAAR
 * @since 13/10/17
 */
public class Arithmetic extends AcqConstraint {

	private final Operator op1, op2; // operators.
	// required visibility to allow exportation
	protected final int cste;

	private static boolean isOperation(Operator operator) {
		return operator.equals(Operator.PL) || operator.equals(Operator.MN);
	}

	protected Arithmetic(String name, AcqVariable[] vars, Operator op1, Operator op2, int cste){

		super(name);

		for(int i=0; i<vars.length;i++)
			this.variables.add(vars[i]);

		this.arity=vars.length;
		this.op1 = op1;
		this.op2 = op2;
		this.cste = cste;
	}

	Arithmetic(AcqVariable var, Operator op, int cste) {
		this(AcqConstraintNames.ARITHM, new AcqVariable[]{var}, op, Operator.NONE, cste);
	}



	Arithmetic(AcqVariable var1, Operator op, AcqVariable var2) {
		this(AcqConstraintNames.ARITHM, new AcqVariable[]{var1,var2}, op, Operator.NONE, 0);
	}



	Arithmetic(AcqVariable var1, Operator op1, AcqVariable var2, Operator op2, int cste) {
		this(AcqConstraintNames.ARITHM, new AcqVariable[]{var1,var2}, op1, op2, cste);	
	}



	@Override
	public AcqConstraint getNegation(){

		if(this.arity==1)
			return new Arithmetic(this.variables.get(0),Operator.getOpposite(op1),cste);
		else{

			assert variables.size()==2;

			if(op1==Operator.PL || op1==Operator.MN){
				return new Arithmetic(this.variables.get(0),op1,this.variables.get(1),Operator.getOpposite(op2),cste);
			}
			else{
				return new Arithmetic(this.variables.get(0),Operator.getOpposite(op1),this.variables.get(1),op2,cste);
			}
		}
	}
	
	public void toChoco(Model model, Arithmetic cstr){
		
		if(cstr.arity==1)
		{
			IntVar x = model.intVar(cstr.variables.get(0).name(), cstr.variables.get(0).min(), cstr.variables.get(0).max());
			model.arithm(x, cstr.op1.toString(), cstr.cste).post();
		}	
		else{						// NL: arity=2
			IntVar x = model.intVar(cstr.variables.get(0).name(), cstr.variables.get(0).min(), cstr.variables.get(0).max());
			IntVar y = model.intVar(cstr.variables.get(1).name(), cstr.variables.get(1).min(), cstr.variables.get(1).max());
				model.arithm(x, cstr.op1.toString(),y,cstr.op2.toString(), cstr.cste).post();
			
		}

	}
}
