/**
 *
 */
package ag;

import org.chocosolver.solver.Model;

import config.Heuristic;
import java.util.ArrayList;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.variables.BoolVar;
import org.chocosolver.solver.variables.IntVar;

/**
 * Class using Choco Solver 4.1.0 as solver
 * 
 * @author LAZAAR
 *
 */
public class ACQ_ChocoSolver extends ACQ_ConstraintSolver {

	/**
	 * Domain of variables (à revoir parce que les variables d'un meme solver peuvent avoir des domaines différents)
	 */
	private ACQ_Domain domain;

	/**
	 * Constructor
	 * 
	 * @param domain Domain of variables
	 */

	public ACQ_ChocoSolver(ACQ_Domain domain) {
		this.domain = domain;
	}

	/**
	 * Generate a query that satisfies networkA but violates at least one constraint of networkB
	 * 
	 * @param network_A Constraint network to satisfy
	 * @param network_B Constraint network to violate
	 * @return Query that satisfies networkA but violates at least one constraint of networkB
	 */
	
	@Override
	public ACQ_Query solve_AnotB(ACQ_Network network_A, ACQ_Network network_B, boolean findc, Heuristic heuristic) {

		System.out.println("ag.ACQ_ChocoSolver.solve_AnotB()");


		Model model = new Model("solveAnotB");

		IntVar[] chocoVars = buildModel(model, network_A, network_B, findc, heuristic);


		Solver solver = model.getSolver();
		
		if(!heuristic.equals(Heuristic.SOL))
				while (solver.solve()) {}
		else solver.solve();
		
		if(solver.getSolutionCount()==0) return new ACQ_Query();
		
		int[] tuple = new int[chocoVars.length];
		for (int i = 0; i < chocoVars.length; i++) {
			tuple[i] = chocoVars[i].getValue();
		}
		ACQ_Query query=new ACQ_Query(network_A.variables, tuple);
		
		return query;
	}
	


	/**
	 * Generate a query that satisfies networkA and at least one constraint of networkB
	 * 
	 * @param networkA Constraint network to satisfy
	 * @param networkB Constraint network where at least one constraint is satisfied
	 * @param heuristic
	 * @return Query that satisfies networkA and at least one constraint of networkB
	 */
	@Override
	public ACQ_Query solve_AnotAllB(ACQ_Network networkA, ACQ_Network networkB, Heuristic heuristic) {

		System.out.println("ag.ACQ_ChocoSolver.solve_AnotAllB()");
		Model model = new Model("solveA");
		IntVar[] chocoVars = buildModel(model, networkA);
		Solver solver = model.getSolver();
		while (solver.solve()) {
			int[] tuple = new int[chocoVars.length];
			for (int i = 0; i < chocoVars.length; i++) {
				tuple[i] = chocoVars[i].getValue();
			}
			ACQ_Query query=new ACQ_Query(networkA.variables, tuple);
			if (networkB.checkNotAllNeg(query)) {
				return query;
			}
		}
		return new ACQ_Query();
	}

	/**
	 * Build the solver model. Add from the specified network, the variables and the constraints to the specified model.
	 * 
	 * @param model Target model to add variables and constraints
	 * @param network Constraint network to modelize
	 * @return An array of all the variables of the model
	 */
	private IntVar[] buildModel(Model model, ACQ_Network network) {
		// declare variables
		ArrayList<IntVar> intVars = new ArrayList<>();
		network.variables.forEach(numvar -> {
			IntVar intVar = model.intVar(NameService.getVarName(numvar), domain.getMin(numvar), domain.getMax(numvar));
			intVars.add(intVar);
		});
		IntVar[] varArray = intVars.toArray(new IntVar[intVars.size()]);
		// add constraints in model
		for (ACQ_IConstraint c : network.getConstraints()) {
			//            c.toChoco(model, c.getVariables(varArray));
			c.toChoco(model, varArray);
		}
		return varArray;

	}

	/**
	 * Build a model (network_A and not network_B) using reified constraints based refutation. 
	 * Assumption: network_A and network_B are expressed on the same set of variables.
	 * 
	 * @param model Target model to add variables and constraints
	 * @param network_A Constraint network to add to *model*
	 * @param network_B Constraint network to add to *model* as refuted one using reified constraints
	 * @param heuristic 
	 * @param findc 
	 * @return varArray model variables
	 * @
	 */
	private IntVar[] buildModel(Model model, ACQ_Network network_A, ACQ_Network network_B, boolean findc, Heuristic heuristic) {

		// declare variables
		ArrayList<IntVar> intVars = new ArrayList<>();
		network_A.variables.forEach(numvar -> {
			IntVar intVar = model.intVar(NameService.getVarName(numvar), domain.getMin(numvar), domain.getMax(numvar));
			intVars.add(intVar);
		});
		IntVar[] varArray = intVars.toArray(new IntVar[intVars.size()]);

		// add constraints in model
		for (ACQ_IConstraint c : network_A.getConstraints()) {
			c.toChoco(model, varArray);
		}

		BoolVar[] reifyArray = model.boolVarArray(network_B.size());
		int i=0;
		for (ACQ_IConstraint c : network_B.getConstraints()) {
			c.toReifiedChoco(model, reifyArray[i], varArray);
			i++;
		}

		IntVar obj_var;
		if(findc)
			obj_var = model.intVar("obj_var", 1, reifyArray.length-1);
		else
			obj_var = model.intVar("obj_var", 0, reifyArray.length-1);

		if(heuristic.equals(Heuristic.MAX))
			model.setObjective(Model.MINIMIZE, obj_var);

		if(heuristic.equals(Heuristic.MIN))
			model.setObjective(Model.MAXIMIZE, obj_var);


		model.sum(reifyArray, "=", obj_var);



		return varArray;

	}
	/**
	 * Ask this solver if there is any solution to the specified network
	 * 
	 * @param learned_network Constraint network to check
	 * @return true if there is at least one solution of the specified network
	 */
	@Override
	public boolean solve(ACQ_Network network) {
		
		Model model = new Model("solveA");
		
		
		buildModel(model, network);
		
		
		return model.getSolver().solve();
	}
}
