/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package ag;

import java.util.Arrays;
import java.util.Iterator;
import org.testng.Assert;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertEquals;

/**
 * Class used to represent constraint network.
 */
public class ACQ_Network implements Iterable<ACQ_IConstraint> {
	/**
	 * Variables of this network
	 */
	protected ACQ_Scope variables;
	/**
	 * Set of constraints of this network
	 */
	protected ACQ_ConstraintSet constraints;
	/**
	 * Empty constructor
	 */
	public ACQ_Network() {
		this(null, null, true);
	}
	/**
	 * Constructor from a specified scope (variables)
	 * 
	 * @param variables Scope that contains variables
	 */
	public ACQ_Network(ACQ_Scope variables) {
		this(variables, null, true);
	}
	
	
	/**
	 * Constructor from a specified set of constraints
	 * 
	 * @param constraints Set of constraints
	 */
	public ACQ_Network(ACQ_ConstraintSet constraints) {
		// variable set is polpulated by used constraints
		this(null, constraints, true);
	}
	/**
	 * Constructor from a specified scope and set of constraints
	 * 
	 * @param variables Scope
	 * @param constraints Set of constraints
	 */
	public ACQ_Network(ACQ_Scope variables, ACQ_ConstraintSet constraints) {
		// as default constraints are only added when all variables are already contained in variable set
		this(variables, constraints, false);
	}
	/**
	 * Constructor from an other network and a specified scope
	 * It takes the constraints from the other network and the 
	 * scope from the specified scope
	 * 
	 * @param other_network Network to take constraints from
	 * @param scope Scope
	 */
	public ACQ_Network(ACQ_Network other_network, ACQ_Scope scope) {
		this(scope, other_network.constraints, false);
	}
	/**
	 * Main constructor of networks
	 * Build a network from the specified scope and the specified set of constraints
	 * 
	 * @param variables Scope 
	 * @param constraints Set of constraints
	 * @param add_variables if false constraints are only added when all variables are already contained in variable set, else it will add the constraint and the missing variables
	 */
	public ACQ_Network(ACQ_Scope variables, ACQ_ConstraintSet constraints, boolean add_variables) {
		if (variables == null) {
			this.variables = new ACQ_Scope();
		} else {
			this.variables = variables;
		}
		if (constraints != null) {
			this.constraints = constraints;
			constraints.forEach((x) -> {
				add(x, add_variables);
			});
		} else {
			this.constraints = new ACQ_ConstraintSet();
		}
	}
	/**
	 * Returns an iterator of the set of constraints of this network
	 * 
	 * @return iterator of the set of constraints
	 */
	@Override
	public Iterator<ACQ_IConstraint> iterator() {
		return constraints.iterator();
	}
	/**
	 * Returns the size of the set of constraints
	 * 
	 * @return amount of constraints
	 */
	public int size() {
		return constraints.size();
	}
	/**
	 * Returns the set of constraints of this network
	 * 
	 * @return set of constraints
	 */
	public ACQ_ConstraintSet getConstraints() {
		return this.constraints;
	}
	/**
	 * Add a specified constraint to the set of this network
	 * If force is set to false, the specified constraint will only be added 
	 * if all the variables of the specified constraint are contained in the variable set of this network
	 * Else it will add the constraint and the missing variables
	 * 
	 * @param cst Constraint to add
	 * @param force Determine the way of adding
	 */
	public final void add(ACQ_IConstraint cst, boolean force) {
		if (force || this.variables.containsAll(cst.getScope())) {
			this.constraints.add(cst);
			this.variables = variables.union(cst.getScope());
		}
	}
	
	/**
	 * Display all the details of the variables of this network
	 * 
	 */
	public void printVariables() {
		for (int var : variables) {
			System.out.println("var " + var + "=" + NameService.getVarName(var));
		}
	}
	/**
	 * Checks if the query violates this network of constraints
	 * 
	 * @param query Example
	 * @return false if query violates at least one constraint of this network
	 */
	boolean check(ACQ_Query query) {
		for (ACQ_IConstraint constraint : constraints) {
			if (query.getScope().containsAll(constraint.getScope()) 
					&& !constraint.checker(constraint.getProjection(query))) {
				return false;
			}
		}
		return true;
	}

	/**************************
	 * checkNotAllNeg  
	 * 
	 * @param query
	 * @return true if the tuple represented by 
	 * query is rejected by at least a constraint of the network
	 * 
	 * @author NADJIB
	 **************************/
	boolean checkNotAllNeg(ACQ_Query query){
		int counter1=0, counter2=0;
		for (ACQ_IConstraint constraint : constraints) {
			if (query.getScope().containsAll(constraint.getScope()))
			{
				counter1++;

				if(constraint.checker(constraint.getProjection(query))) 
					counter2++;

			}

		}
		return (counter2 != 0) && (counter2 != counter1);
	}

	@Override
	public String toString() {
		String s= "network var=" + variables.size() + " cst=" + constraints.size()+"\n";

		s+="-------------------------\n";
		s+="CONSTRAINTS:\n";
		for(ACQ_IConstraint c: this.constraints)
			s+= c.toString()+"\n";
		s+="-------------------------\n";

		return s;
	}
}
