/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package ag;

import acqConstraints.Operator;
import org.chocosolver.solver.Model;
import org.chocosolver.solver.variables.BoolVar;
import org.chocosolver.solver.variables.IntVar;

/**
 *	Constraint between two variables. It can either be a straight comparison between these two
 *  For example : X >= Y or X != Y
 *  Or it can be an operation followed by a comparison using a constant
 *  For example : X + 5 <= Y or X - Y = 19
 *	
 * @author agutierr
 */
public class BinaryArithmetic extends BinaryConstraint {
	/**
	 * Operator(s) used into this constraint. It must have at least one operator, 
	 * the second is optional depending on the constraint we want to build.
	 */
	private final Operator op1, op2;

	// required visibility to allow exportation
	/**
	 * Optional constant value to allow more comparison for the constraint
	 */
	protected final int cste;

	/**
	 * Constructor for a constraint between two variables. 
	 * 
	 * @param name Name of this constraint
	 * @param var1 Variable of this constraint
	 * @param op Operator of this constraint
	 * @param var2 Variable of this constraint
	 */
	public BinaryArithmetic(String name, int var1, Operator op, int var2) {
		super(name, var1, var2);
		this.op1 = op;
		this.op2 = Operator.NONE;
		this.cste = -1; // not used
	}

	/**
	 * Constructor for a constraint involving two variables and a constant value
	 * 
	 * @param name Name of this constraint
	 * @param var1 Variable of this constraint
	 * @param op1 Operator of this constraint
	 * @param var2 Variable of this constraint 
	 * @param op2 Operator of this constraint
	 * @param cste Constant value for this constraint
	 */
	public BinaryArithmetic(String name, int var1, Operator op1, int var2, Operator op2, int cste) {
		super(name, var1, var2);
		this.op1 = op1;
		this.op2 = op2;
		this.cste = cste;
	}

	/**
	 * Checks if this constraint has two operators
	 * 
	 * @return true if this constraint has two operators
	 */
	private boolean hasOperation() {
		return op2 != Operator.NONE;
	}

	/**
	 * Returns a new BinaryArithmetic constraint which is the negation of this constraint
	 * By instance, a constraint with "=" as operator will return a new constraint with the 
	 * same variables but with "!=" as operator 
	 * 
	 * @return A new BinaryArithmetic constraint, negation of this constraint
	 */
	@Override
	public BinaryArithmetic getNegation() {
		if (op1 == Operator.PL || op1 == Operator.MN) {
			return new BinaryArithmetic("not_" + getName(), getScope().getFirst(), op1, getScope().getSecond(), Operator.getOpposite(op2), cste);
		} else {
			return new BinaryArithmetic("not_" + getName(), getScope().getFirst(), Operator.getOpposite(op1), getScope().getSecond(), op2, cste);
		}

	}

	/**
	 * Add this constraint to the specified model (a choco solver model in this case)
	 * 
	 * @param model Model to add this constraint to
	 * @param intVars Variables of the model involved in this constraint
	 * 
	 */
	@Override
	public void toChoco(Model model, IntVar... intVars) {

		if (hasOperation()) {
	//		if(this.getVariables().length==intVars.length)
	//			model.arithm(intVars[0], op1.toString(), intVars[1], op2.toString(), cste).post();
	//		else
				model.arithm(intVars[this.getVariables()[0]], op1.toString(), intVars[this.getVariables()[1]], op2.toString(), cste).post();
		} else {
	//		if(this.getVariables().length==intVars.length)
	//			model.arithm(intVars[0], op1.toString(), intVars[1]).post();
	//		else			
				model.arithm(intVars[this.getVariables()[0]], op1.toString(), intVars[this.getVariables()[1]]).post();
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + cste;
		result = prime * result + ((op1 == null) ? 0 : op1.hashCode());
		result = prime * result + ((op2 == null) ? 0 : op2.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {


		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		BinaryArithmetic other = (BinaryArithmetic) obj;
		if (cste != other.cste)
			return false;
		if (op1 != other.op1)
			return false;
		if (op2 != other.op2)
			return false;
		if (this.getVariables()[0] != other.getVariables()[0])
			return false;
		if (this.getVariables()[1] != other.getVariables()[1])
			return false;

		return true;
	}

	/**
	 * Add this constraint to the specified model (a choco solver model in this case)
	 * 
	 * @param model Model to add this constraint to
	 * @param intVars Variables of the model involved in this constraint
	 * 
	 */
	@Override
	public void toReifiedChoco(Model model, BoolVar b,IntVar... intVars) {

		int left=0, right=1;
		if(intVars.length>this.getVariables().length)
		{
			left= this.getVariables()[0];
			right=this.getVariables()[1];
		}
		if (hasOperation()) {
			model.arithm(intVars[left], op1.toString(), intVars[right], op2.toString(), cste).reifyWith(b);
		} else {
			model.arithm(intVars[left], op1.toString(), intVars[right]).reifyWith(b);
		}

	}
	/**
	 * Checks this constraint for a specified set of values
	 * 
	 * @param value1 Value of the first variable of this constraint
	 * @param value2 Value of the second variable of this constraint
	 * @return true if this constraint is satisfied for the specified set of values
	 */
	@Override
	protected boolean check(int value1, int value2) {
		int val1, val2;
		Operator op;
		if(op2 == Operator.NONE) {
			val1 = value1;
			val2 = value2;
			op = op1;
		}
		else {
			if (isOperation(op1)) {

				if (op1 == Operator.PL) {
					val1 = value1 + value2;
				} else {
					val1 = value1 - value2;
				}
				val2 = cste;
				op = op2;
			} else {
				if (op2 == Operator.PL) {
					val2 = value2 + cste;
				} else {
					val2 = value2 - cste;
				}
				val1 = value1;
				op = op1;
			}
		}	
		switch (op) {
		case EQ:
			return val1 == val2;
		case NQ:
			return val1 != val2;
		case GT:
			return val1 > val2;
		case GE:
			return val1 >= val2;
		case LT:
			return val1 < val2;
		case LE:
			return val1 <= val2;
		}
		return false;
	}

	/**
	 * Checks if the specified operator is an operation operator
	 * 
	 * @param operator Operator to verify
	 * @return true if the specified operator is Operator.PL or Operator.MN
	 */
	private static boolean isOperation(Operator operator) {
		return operator.equals(Operator.PL) || operator.equals(Operator.MN);
	}

}
