package algo;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import acqConstraints.AcqConstraint;
import cp.ConstraintSolver;
import machineLearning.Bias;
import machineLearning.Query;
import machineLearning.Learner;
import tools.AcqCollection;
import vocabulary.AcqNetwork;
import vocabulary.AcqVariable;
import config.Parameters;
import config.Heuristic;
import config.Query_type;

public class QUACQ implements Acquisition{

	private AcqNetwork learned_network;
	private ConstraintSolver solver;
	private Bias bias;
	private Learner learner;
	private Heuristic heuristic;


	public QUACQ(){

		//NL: config part
		this.heuristic=Parameters.get_heuristic();
		this.solver=Parameters.get_solver();
		this.bias= Parameters.get_bias();
		this.learner= Parameters.get_lerner();
		this.learned_network= new AcqNetwork(bias.getVars());
		//////////////////////////////////

	}


	@Override
	public boolean quacq_process() {

		boolean convergence=false;
		boolean collapse=false;


		assert(learned_network.size()==0);
		while(!convergence && !collapse){


			if(!solver.solve(learned_network) && !solver.timeout_reached()){
				collapse=true;
			}


			Query membership_query=query_gen(learned_network, bias, bias.getVars(), Query_type.MQ, heuristic );

			if(membership_query.isEmpty())
				convergence=true;

			else if(learner.ask(membership_query)){
				bias.reduce(membership_query);
			}else{
				AcqConstraint cst= findC(findScope(membership_query, (List<AcqVariable>) bias.getVars(), 
						new ArrayList<AcqVariable>(), false),membership_query);	
				if(cst.isEmpty())
					collapse=true;
				else
					learned_network.add(cst);
			}
		}

		return !collapse;

	}



	@Override
	/******************************************************************
	 * findC (classic findC of IJCAI13)
	 * 
	 * @date 03-10-17
	 * @author LAZAAR
	 *******************************************************************/
	public AcqConstraint findC(List<AcqVariable> scope, Query e) {

		AcqNetwork learned_network_y=new AcqNetwork(learned_network,scope);
		SortedSet<AcqVariable> scope_set =new TreeSet<AcqVariable>(scope);

		Bias bias_y=new Bias(bias.getProjection(scope), scope_set);
		List<AcqConstraint> temp_kappa =new ArrayList<AcqConstraint>();

		assert scope.size() > 1;



		List<AcqConstraint> candidates= new ArrayList<AcqConstraint>();  //NL:  candidates = delta in IJCAI13
		candidates.addAll(bias_y.getConstraints());

		assert(candidates.size()==bias_y.size());


		candidates.retainAll(bias.getKappa(e));


		while(true) {

			if(candidates.isEmpty())
				return null;
			//TODO check if returning null or empty constraint (NL: 03-10-17)			

			int temp=candidates.size();

			Query partial_findC_query= query_gen(learned_network_y, new AcqNetwork(scope_set,candidates),scope_set, Query_type.findc, Heuristic.SOL);

			assert(candidates.size()==temp);

			if (partial_findC_query.isEmpty()) {

				bias.reduce(candidates);

				return candidates.get(0);

			}

			temp_kappa=(List<AcqConstraint>) bias.getKappa(partial_findC_query);

			if(temp_kappa.isEmpty())
				throw new RuntimeException("Collapse state");


			learner.asked_query(partial_findC_query);

			if(!partial_findC_query.isClassified())
			{
				learner.ask(partial_findC_query);
				learner.memory_up(partial_findC_query);
			}


			if (partial_findC_query.isPositive()) {

				bias.reduce(temp_kappa);			// NL: kappa of partial_findC_query

				candidates.removeAll(temp_kappa);

			}
			else{
				candidates=(List<AcqConstraint>) AcqCollection.intersection(candidates,temp_kappa);		//TESTME

			}
		}

	}


	/****************************************************************
	 *  FindScope procedure
	 * @param negative_example : a complete negative example
	 * @param X : problem variables (and/or) foreground variables
	 * @param Bgd : background variables
	 * @param mutex : boolean mutex
	 * @return  variable scope
	 * 
	 * @author LAZAAR
	 ****************************************************************/

	@Override
	public List<AcqVariable> findScope(Query negative_example,
			List<AcqVariable> X, List<AcqVariable> Bgd, boolean mutex) {


		SortedSet<AcqConstraint> temp_kappa=new TreeSet<AcqConstraint>();


		if (mutex && Bgd.size()>= bias.getMinArity()  ) {				//TESTME if minArity has the good value !!

			Query query_bgd = negative_example.getProjection(Bgd);      // projection e|Bgd

			temp_kappa= (SortedSet<AcqConstraint>) bias.getKappa(query_bgd); 


			learner.asked_query(query_bgd);

			if(!query_bgd.isClassified())
			{

				learner.ask(query_bgd);
				learner.memory_up(query_bgd);

			}

			if(!query_bgd.isPositive()) 			//NL: negative
				return new ArrayList<AcqVariable>();		//NL: return emptyset
			else 
				bias.reduce(temp_kappa);

		}


		if(X.size()==1) return (List<AcqVariable>) X;

		List<AcqVariable> X1;
		List<AcqVariable> X2;
		List<AcqVariable> S1= new ArrayList<AcqVariable>();
		List<AcqVariable> S2= new ArrayList<AcqVariable>();
		List<AcqVariable> Bgd_union_X2 = new ArrayList<AcqVariable>();
		List<AcqVariable> Bgd_union_S1 = new ArrayList<AcqVariable>();
		List<AcqVariable> S1_union_S2 = new ArrayList<AcqVariable>();


		X1 = new ArrayList<AcqVariable>(X.subList(0, X.size()/2));		//TESTME
		X2 = new ArrayList<AcqVariable>(X.subList(X.size()/2, X.size()));
		//NL: different splitting manners can be defined here!

		Bgd_union_X2.addAll(Bgd);
		Bgd_union_X2.addAll(X2);

		S1=findScope(negative_example,X1,Bgd_union_X2,true);    // NL: First recursive call of findScope

		Bgd_union_S1.addAll(Bgd);
		Bgd_union_S1.addAll(S1);
		mutex= !S1.isEmpty();
		S2=findScope(negative_example,X2,Bgd_union_S1,mutex);    // NL: Second recursive call of findScope

		S1_union_S2.addAll(S1);
		S1_union_S2.addAll(S2);

		return S1_union_S2;
	}


	/****************************************************************************
	 * query_gen
	 * 
	 * @param network1
	 * @param network2
	 * @param scope
	 * @param type
	 * @return Query
	 * @author LAZAAR
	 * @date 03-10-2017
	 * 
	 * get a query of type "type" on scope "scope" s.t., network1 and not network2
	 *****************************************************************************/

	@Override
	public Query query_gen(AcqNetwork network1, AcqNetwork network2, SortedSet<AcqVariable> scope, Query_type type, Heuristic h) {

		switch(type){
		case MQ:
			return solver.solve_AnotB(network1,network2,heuristic);

		}

		//notAllViolated in solver for findC query generation
		return null;
	}

}
